import { Show, createEffect, createMemo, createSignal } from "solid-js";
import { createStore, reconcile } from "solid-js/store";
import { PixelPerfectCanvas } from "@shadryx/pptk/solid";
import { Client } from "boardgame.io/client";
import { Local, SocketIO } from "boardgame.io/multiplayer";

import { SOCKETIO_SERVER } from "../consts.js";
import clone from "../clone.js";
import * as game from "../game.js";

import Toolbar, { Tool } from "./Toolbar.jsx";
import Status from "./Status.tsx";

import classes from "./Game.module.css";
import * as draw from "./Game/draw.ts";
import useAnimation from "./Game/useAnimation.ts";
import useDrawBoard from "./Game/useDrawBoard.ts";
import useBoardMouse from "./Game/useBoardMouse.ts";

export type GameProps = ({
    playerID: string,
    solo: true,
    matchID?: undefined,
} | {
    playerID: string,
    credentials: string,
    solo?: false,
    matchID: string,
} | {
    playerID?: undefined,
    solo?: false,
    matchID: string,
}) & {
    debug?: boolean
};

export default function Game(props: GameProps) {
    const [canvas, setCanvas] = createSignal<HTMLCanvasElement>();
    const [width, setWidth] = createSignal(0);
    const [height, setHeight] = createSignal(0);

    const client = Client({
        game: game.AcrossTheHex,
        multiplayer: props.solo ? Local() : SocketIO({
            server: SOCKETIO_SERVER
        }),
        debug: props.debug ?? false,
        ...(props.matchID ? { matchID: props.matchID } : {}),
        ...(props.playerID ? { playerID: props.playerID } : {}),
        ...("credentials" in props ? { credentials: props.credentials } : {}),
    });
    const [gameOver, setGameOver] = createSignal(client.getState()?.ctx.gameover ?? false);
    const [state, setState] = createStore<game.State>(clone(client.getState()?.G) ?? {
        resources: {},
        cells: {},
        moves: {}
    });
    const [turn, setTurn] = createSignal(1);
    const [soloTurn, setSoloTurn] = createSignal<string>("");
    const [stage, setStage] = createSignal<string>();

    const {
        start: startAnimation,
        running: animationRunning,
        get: getAnimationState,
    } = useAnimation(state, 333);

    const [hoveredCell, setHoveredCell] = createSignal<[number, number] | null>(null);
    const [selectedTool, setSelectedTool] = createSignal<Tool>({
        type: "placeBuilding",
        building: "road"
    });

    const moves = () => client.moves as game.Moves;

    const getTransform = createMemo(
        () => draw.getTransform(state.cells, width(), height()),
        {
            scale: 1,
            sx: 0,
            sy: 0,
        },
        {
            equals(prev, next) {
                return prev.scale === next.scale && prev.sx === next.sx && prev.sy === next.sy
            }
        }
    );

    const drawBoard = useDrawBoard({
        canvas,
        state,
        getTransform,
        hoveredCell,
        gameOver,
        selectedTool,
        getAnimationState,
        playerID: props.playerID,
        targetTiles: createMemo(() => {
            const tool = selectedTool();
            if (tool.type !== "attack" || !tool.selected || !props.playerID) return [];

            const building = game.getCell(state.cells, tool.selected.x, tool.selected.y);
            if (building?.owner !== props.playerID) return [];
            const buildingRules = game.Buildings[building.building];
            if (!("attack" in buildingRules)) return [];

            return buildingRules.attack.targets(state.cells, tool.selected.x, tool.selected.y)
                .filter(([x, y]) => game.cellIn(state.cells, x, y));
        }),
    });

    useBoardMouse({
        canvas,
        state,
        setHoveredCell,
        getTransform,
        clickTile(x: number, y: number) {
            if (!props.playerID) return;
            if (animationRunning() || gameOver()) return;

            const tool = selectedTool();

            if (tool.type === "placeBuilding") {
                moves().placeBuilding?.(x, y, tool.building);
            } else if (tool.type === "attack") {
                if (tool.selected) {
                    if (game.canAttack(
                        state.cells,
                        game.remainingResources(state.cells, state.resources, state.moves, props.playerID),
                        props.playerID,
                        tool.selected.x,
                        tool.selected.y,
                        x,
                        y
                    )) {
                        moves().attack?.(tool.selected.x, tool.selected.y, x, y);
                        setSelectedTool({
                            type: "attack",
                            selected: null
                        });
                    }
                } else {
                    setSelectedTool({
                        type: "attack",
                        selected: {
                            x,
                            y
                        }
                    });
                }
            }
        },
    });

    client.start();
    client.subscribe((state) => {
        if (state) {
            if (state.ctx.turn > turn()) {
                setTurn(state.ctx.turn);
                startAnimation();
            }

            setGameOver(state.ctx.gameover ?? false);

            // For some ungodly reason, I need to clone the state, because otherwise the frozenness sporadically gets
            // saved in the store, triggering various issues later down the line.
            // My guess for as to why is that `state.G` can be mutated and some of its properties are passed to Object.freeze
            setState(reconcile(clone(state.G)));
            if (props.playerID) {
                setStage(state.ctx.activePlayers?.[props.playerID] ?? "connecting");
            }
            if (props.solo) {
                const activePlayers = Object.entries(state.ctx.activePlayers ?? {})
                    .flatMap(([player, state]) => state === "prepare" ? [player] : [])
                    .sort((a, b) => +b - +a);
                setSoloTurn(activePlayers[activePlayers.length - 1] ?? "0");
            }
            console.log(state.ctx.activePlayers);
        }
    });

    createEffect(() => {
        drawBoard();
    });

    return <div class={classes.game}>
        <Show when={props.solo === true ? soloTurn() === props.playerID : true}>
            <div class={classes.board}>
                <PixelPerfectCanvas
                    onAttach={setCanvas}
                    onResize={(_, width, height) => {
                        setWidth(width);
                        setHeight(height);
                        drawBoard();
                    }}
                    style={{
                        width: "min(900px, 50vw, 80vh - 13em)",
                        height: "min(900px, 50vw, 80vh - 13em)",
                    }}
                />
                <div class={classes.turn}>
                    Turn {turn()}{gameOver() ? "- Game over" : ""}
                </div>
            </div>
            <Show when={props.matchID}>
                {(matchID) => <div class={classes.players}><Status gameName={game.AcrossTheHex.name!} matchID={matchID()} /></div>}
            </Show>
            <Show when={props.playerID}>
                {(playerID) => {
                    return <div class={classes.toolbar}><Toolbar
                        playerID={playerID()}
                        selectedTool={selectedTool}
                        setSelectedTool={setSelectedTool}
                        resources={() => "resources" in state ? game.remainingResources(state.cells, state.resources, state.moves, playerID()) : 0}
                        resourceGain={() => "cells" in state ? game.getResourceGain(state.cells, playerID()) : 0}
                        setReady={() => {
                            if (animationRunning()) return;
                            moves().setReady?.();
                            if (selectedTool().type === "attack") {
                                setSelectedTool({
                                    type: "attack",
                                    selected: null
                                });
                            }
                        }}
                        ready={() => stage() === "ready"}
                    /></div>;
                }}
            </Show>
        </Show>
    </div>;
}
